// vim: set ts=2 sw=2 et:

#include <iostream>
#include <queue>
#include <map>
#include <unordered_map>
#include <vector>
#include <cassert>
#include <cstdint>
#include <algorithm>
#include <memory.h>
#include "folly/FBString.h"
#include "folly/FBVector.h"
#include "folly/experimental/EventCount.h"

#include "reader.h"
#include "protocol.h"
#include "parser.h"
#include "relation.h"
#include "transactor.h"
#include "validator.h"
#include "prevalidator.h"

using namespace std;

#ifdef _PROFILE
TimestampStack timestampStackMainThread;
ProfilingStats timeSinglethreaded;
TimestampInstant startTimeMainThread;
TimestampUnit profilingOverhead;
#endif

const int Transactor::THREAD_COUNT = 2;
const int Prevalidator::THREAD_COUNT = 5;
const int Validator::THREAD_COUNT = 16;

Relation *relations;
Relation *relationsEnd;

static uint64_t firstUnknownValidationId;
folly::LifoSem flushReadySemaphore;
folly::EventCount relationReadyEventCount;
std::atomic<uint64_t> flushCounter;

struct MainThreadDelegate;
static folly::fbvector<ValidationResult> prevValidations;
static Transactor::ThreadContext *transactorThreadContexts;
static Prevalidator::ThreadContext *prevalidatorThreadContexts;
static Validator::ThreadContext *validatorThreadContexts;
static Parser<MainThreadDelegate> *mainThreadParser;

struct SchedulingPlan
{
  folly::fbvector<pair<uint64_t, folly::fbvector<uint32_t>>> costs;
  uint64_t costSum;

  inline SchedulingPlan(uint64_t singleCost, uint32_t index)
  {
    costs.emplace_back(singleCost, folly::fbvector<uint32_t>({ index }));
    for (int i = 1;i < Transactor::THREAD_COUNT;i++)
      costs.emplace_back(0, folly::fbvector<uint32_t>());
    costSum = singleCost;
  }

  inline bool operator < (const SchedulingPlan &other) const
  {
    return costSum < other.costSum;
  }

  void merge(const SchedulingPlan &other)
  {
    for (size_t i = 0;i < costs.size();i++)
    {
      int ri = costs.size() - i - 1;
      costs[i].first += other.costs[ri].first;
      costs[i].second.insert(costs[i].second.end(), other.costs[ri].second.begin(), other.costs[ri].second.end());
    }

    sort(costs.begin(), costs.end(), greater<pair<uint64_t, folly::fbvector<uint32_t>>>());
    costSum = 0;
    for (size_t i = 0;i < costs.size();i++)
      costSum += (costs[i].first -= costs[costs.size() - 1].first);
  }

  void debug() const
  {
    for (int i = 0;i < Transactor::THREAD_COUNT;i++)
    {
      fprintf(stderr, "Thread %d: ", i);
      uint64_t cost = 0;
      for (auto val : costs[i].second)
      {
        fprintf(stderr, "%d ", val);
        cost += relations[val].approximateCost;
      }
      fprintf(stderr, " ==> cost %lu\n", cost);
    }
  }
};

// Karmarkar-Karp
static void rescheduleThreads()
{
  priority_queue<SchedulingPlan> plans;
  for (int i = 0;i < relationsEnd - relations;i++)
    plans.emplace(relations[i].approximateCost, i);
  while (plans.size() != 1)
  {
    SchedulingPlan p = plans.top();
    plans.pop();
    const SchedulingPlan &q = plans.top();
    p.merge(q);
    plans.pop();
    plans.emplace(move(p));
  }
  SchedulingPlan root = plans.top();
  for (int i = 0;i < Transactor::THREAD_COUNT;i++)
    for (auto val : root.costs[i].second)
      relations[val].ownerThreadIndex = i;
}

static void startThreads()
{
  flushCounter = Transactor::THREAD_COUNT + Prevalidator::THREAD_COUNT;
  for (int i = 0;i < Transactor::THREAD_COUNT;i++)
  {
    auto ctxPtr = transactorThreadContexts + i;
    thread([ctxPtr]() { ctxPtr->mainLoop(); }).detach();
  }
  for (int i = 0;i < Prevalidator::THREAD_COUNT;i++)
  {
    auto ctxPtr = prevalidatorThreadContexts + i;
    thread([ctxPtr]() { ctxPtr->mainLoop(); }).detach();
  }
  for (int i = 0;i < Validator::THREAD_COUNT;i++)
  {
    auto ctxPtr = validatorThreadContexts + i;
    thread([ctxPtr]() { ctxPtr->mainLoop(); }).detach();
  }
}

// from testdriver.cpp
static void readAll(void *ptr, size_t bytes)
{
  while (bytes)
  {
    int res = read(STDIN_FILENO, ptr, bytes);
    if (res < 1)
      exit(1);
    ptr = static_cast<char*>(ptr) + res;
    bytes -= res;
  }
}

static void doDefineSchema()
{
  MessageHead messageHead;
  readAll(reinterpret_cast<void *>(&messageHead), sizeof(MessageHead));
  char *buf = new char[messageHead.messageLen];
  readAll(reinterpret_cast<void *>(buf), messageHead.messageLen);
  DefineSchema *d = reinterpret_cast<DefineSchema *>(buf);

  auto relationBuf = new char[sizeof(Relation) * d->relationCount];
  relations = reinterpret_cast<Relation *>(relationBuf);
  for (decltype(d->relationCount) i = 0;i < d->relationCount;i++)
    new (relations + i) Relation(d->columnCounts[i], i % Transactor::THREAD_COUNT);
  relationsEnd = relations + d->relationCount;

  auto transactorThreadContextBuf = new char[sizeof(Transactor::ThreadContext) * Transactor::THREAD_COUNT];
  transactorThreadContexts = reinterpret_cast<Transactor::ThreadContext *>(transactorThreadContextBuf);
  for (int i = 0;i < Transactor::THREAD_COUNT;i++)
    new (transactorThreadContexts + i) Transactor::ThreadContext(i);

  auto prevalidatorThreadContextBuf = new char[sizeof(Prevalidator::ThreadContext) * Prevalidator::THREAD_COUNT];
  prevalidatorThreadContexts = reinterpret_cast<Prevalidator::ThreadContext *>(prevalidatorThreadContextBuf);
  for (int i = 0;i < Prevalidator::THREAD_COUNT;i++)
    new (prevalidatorThreadContexts + i) Prevalidator::ThreadContext(i);

  auto validatorThreadContextBuf = new char[sizeof(Validator::ThreadContext) * Validator::THREAD_COUNT];
  validatorThreadContexts = reinterpret_cast<Validator::ThreadContext *>(validatorThreadContextBuf);
  for (int i = 0;i < Validator::THREAD_COUNT;i++)
    new (validatorThreadContexts + i) Validator::ThreadContext();

  delete[] buf;
}

static void processFlush(const Flush *f)
{
  flushReadySemaphore.wait();

  PROFILE_START(timestampStackMainThread, timeSinglethreaded);

  // flush asap so that we can heat up the read buffer
  auto prevSize = prevValidations.size();
  auto firstKnownValidationId = firstUnknownValidationId - prevSize;
  for (int i = 0;i < Validator::THREAD_COUNT;i++)
  {
    auto &ctxResults = validatorThreadContexts[i].validationResults;
    auto newSize = ctxResults.size();
    if (prevValidations.size() < prevSize + newSize)
      prevValidations.resize(prevSize + newSize);
    for (size_t j = 0;j < newSize;j++)
      prevValidations[prevSize + j] = max(prevValidations[prevSize + j], ctxResults[j]);
    ctxResults.clear();
  }

  if (f != nullptr)
  {
    auto id = firstKnownValidationId;
    size_t idx = 0;
    folly::fbstring output;
    for (;id <= f->validationId && idx < prevValidations.size();id++, idx++)
    {
      switch (prevValidations[idx])
      {
        case VALIDATION_FALSE: output += '0'; break;
        case VALIDATION_TRUE: output += '1'; break;
        case VALIDATION_NOT_EXISTS: break;
      }
    }

    fwrite(output.c_str(), 1, output.size(), stdout);
    fflush(stdout);
    // even if we flush, transactors & validators are sleeping
    firstUnknownValidationId = firstKnownValidationId + prevValidations.size();
    prevValidations.erase(prevValidations.begin(), prevValidations.begin() + idx);
    for (;prevValidations.size() > 0 && prevValidations.back() == VALIDATION_NOT_EXISTS;)
      prevValidations.pop_back();
  }

  readBufferFlushConsumerComplete(mainThreadParser->readBufferReadPos);

  for (Relation *ptr = relations;ptr != relationsEnd;ptr++)
  {
    ptr->readyForValidation = false;
    memcpy(ptr->historicMinValue, ptr->minValue, sizeof(ptr->minValue[0]) * ptr->columnCount);
    memcpy(ptr->historicMaxValue, ptr->maxValue, sizeof(ptr->maxValue[0]) * ptr->columnCount);
    if (!ptr->xactHistory.empty())
      ptr->firstUnknownXactId = ptr->xactHistory.back().xactId; // xact ids are not unique
  }

  rescheduleThreads();

  for (int i = 0;i < Validator::THREAD_COUNT;i++)
    validatorThreadContexts[i].firstUnknownValidationId = firstUnknownValidationId;

  flushCounter = Transactor::THREAD_COUNT + Prevalidator::THREAD_COUNT;
  for (int i = 0;i < Transactor::THREAD_COUNT;i++)
    transactorThreadContexts[i].flushCompleteSemaphore.post();
  for (int i = 0;i < Prevalidator::THREAD_COUNT;i++)
    prevalidatorThreadContexts[i].flushCompleteSemaphore.post();
}

static void processDone()
{
#ifdef _PROFILE
  {
    cerr << "Time spent in main thread solely: " << timeSinglethreaded.totalTimeToSeconds() << "s" << endl;
    cerr << "==== Transactor threads ====" << endl;
    for (int i = 0;i < Transactor::THREAD_COUNT;i++)
      cerr << "- Thread " << i << endl << transactorThreadContexts[i].profilingStatsToString() << endl;
    cerr << "==== Prevalidator threads ====" << endl;
    for (int i = 0;i < Prevalidator::THREAD_COUNT;i++)
      cerr << "- Thread " << i << endl << prevalidatorThreadContexts[i].profilingStatsToString() << endl;
    cerr << "==== Validator threads ====" << endl;
    for (int i = 0;i < Validator::THREAD_COUNT;i++)
      cerr << "- Thread " << i << endl << validatorThreadContexts[i].profilingStatsToString() << endl;
  }
#endif
  quick_exit(0);
}

struct MainThreadDelegate
{
  inline void processDone() { ::processDone(); }
  inline void processTransaction(Transaction *t) {  }
  inline void processValidationQueries(ValidationQueries *v) {  }
  inline void processFlush(Flush *f) { ::processFlush(f); }
  inline void processForget(Forget *f) {  }
};

int main()
{
#ifdef _PROFILE
  const int PROFILING_OVERHEAD_MEASUREMENTS = 1000000;
  TimestampUnit profilingOverhead = TimestampUnit::zero();
  TimestampInstant profilingOverheadLastInstant = TimestampClock::now();
  for (int i = 0;i < PROFILING_OVERHEAD_MEASUREMENTS;i++)
  {
    auto next = TimestampClock::now();
    profilingOverhead += (next - profilingOverheadLastInstant);
    profilingOverheadLastInstant = next;
  }
  ::profilingOverhead = profilingOverhead / PROFILING_OVERHEAD_MEASUREMENTS;

  startTimeMainThread = TimestampClock::now();
#endif

  doDefineSchema();

  startReaderThread();

  startThreads();

  MainThreadDelegate mainThreadDelegate;
  mainThreadParser = new Parser<MainThreadDelegate>(mainThreadDelegate);
  mainThreadParser->parseLoop();

  return 0;
}

