#pragma once

#include <cstdio>
#include <cstdint>
#include <vector>

struct Uint64MemoryPool
{
    const static uint32_t INDEX_BITS = 5;
    const static uint32_t UNIT_SIZE_ALIGNMENT = (1 << (INDEX_BITS - 3));
    const static uint32_t BLOCKS = (1 << INDEX_BITS);
    const static uintptr_t BLOCK_BITMASK = ((1 << INDEX_BITS) - 1);
    const static uintptr_t ALIGN_BITMASK = ~BLOCK_BITMASK;
    const static uint32_t INITIAL_CAPACITY = 256;

    // 3/2 = 1.5
    const static int GROWTH_FACTOR_NUM = 3;
    const static int GROWTH_FACTOR_DEN = 2;

    struct Block
    {
        uint64_t *start;
        uint64_t *body;
        size_t elements;
        size_t bitmapCount;
    };

    Uint64MemoryPool(size_t _unitSize)
        : unitSize((_unitSize + UNIT_SIZE_ALIGNMENT - 1) / UNIT_SIZE_ALIGNMENT * UNIT_SIZE_ALIGNMENT)
    {
        static_assert(INITIAL_CAPACITY % 256 == 0, "should be multiple of 256");
        blocks[0].elements = INITIAL_CAPACITY;
        blocks[0].bitmapCount = INITIAL_CAPACITY / 64;
        if (posix_memalign(reinterpret_cast<void **>(&blocks[0].start), BLOCKS, 8 * (INITIAL_CAPACITY / 64 + INITIAL_CAPACITY * unitSize)))
            perror("Uint64MemoryPool posix_memalign()");
        blocks[0].start[0] = -1ULL;
        blocks[0].body = blocks[0].start + INITIAL_CAPACITY / 64;

        lastBlock = 0;
        lastBlockBitsetCount = 0;
        blockCount = 1;
    }

    /*~Uint64MemoryPool()
    {
        for (uint32_t i = 0;i < blockCount;i++)
            free(blocks[i].start);
    }*/

    inline static uint64_t *align (uint64_t *ptr)
    {
        return reinterpret_cast<uint64_t *>(reinterpret_cast<uintptr_t>(ptr) & ALIGN_BITMASK);
    }

    inline uint64_t *allocate()
    {
        for (uint32_t index = lastBlock;;)
        {
            size_t bitmapCount = blocks[index].bitmapCount;
            for (size_t i = lastBlockBitsetCount;;)
            {
                uint64_t *bitmap = blocks[index].start + i;
                if (!*bitmap)
                {
                    i++;
                    if (i == bitmapCount)
                        break;
                    continue;
                }
                int bitmapIndex = __builtin_ctzll(*bitmap);
                *bitmap ^= (1ULL << bitmapIndex);
                uintptr_t addr = reinterpret_cast<uintptr_t>(blocks[index].body) + (i * 64 + bitmapIndex) * unitSize * sizeof(uint64_t);
                lastBlock = index;
                lastBlockBitsetCount = i;
                return reinterpret_cast<uint64_t *>(addr + index);
            }

            index++;
            lastBlockBitsetCount = 0;
            if (index == blockCount)
                index = 0;
            if (index == lastBlock)
                break;
        }

        allocateBlock();
        uint64_t *bitmap = blocks[lastBlock].start;
        *bitmap ^= 1;
        lastBlockBitsetCount = 0;
        return blocks[lastBlock].body;
    }

    inline void deallocate(uint64_t *address)
    {
        uintptr_t addressInt = reinterpret_cast<uintptr_t>(address);
        uint64_t blockIndex = (addressInt & BLOCK_BITMASK);
        size_t bitmapCount = blocks[blockIndex].bitmapCount;
        uint64_t elementIndex = ((addressInt & ALIGN_BITMASK) - reinterpret_cast<uintptr_t>(blocks[blockIndex].start) - bitmapCount * sizeof(uint64_t)) / unitSize;
        uint64_t bitmapGroup = elementIndex >> 6;
        bitmapGroup ^= (1ULL << (elementIndex & 63));
    }

    inline void allocateBlock()
    {
        if (blockCount == BLOCKS)
        {
            fprintf(stderr, "Cannot allocate more blocks\n");
            exit(1);
        }

        blocks[blockCount].elements = (blocks[blockCount - 1].elements * GROWTH_FACTOR_NUM + GROWTH_FACTOR_DEN - 1) / GROWTH_FACTOR_DEN;
        blocks[blockCount].bitmapCount = (blocks[blockCount].elements + 63) >> 6;
        auto alignedBitsetCount = (blocks[blockCount].bitmapCount + UNIT_SIZE_ALIGNMENT - 1) / UNIT_SIZE_ALIGNMENT * UNIT_SIZE_ALIGNMENT;
        if (posix_memalign(reinterpret_cast<void **>(&blocks[blockCount].start), BLOCKS, 8 * (alignedBitsetCount + blocks[blockCount].elements * unitSize)))
            perror("Uint64MemoryPool posix_memalign()");
        blocks[blockCount].body = blocks[blockCount].start + alignedBitsetCount;
        for (size_t i = 0;i != blocks[blockCount].bitmapCount;i++)
            blocks[blockCount].start[i] = -1ULL;
        blocks[blockCount].start[blocks[blockCount].bitmapCount - 1] = (1ULL << (blocks[blockCount].elements & 63)) - 1;

        lastBlock = blockCount;
        blockCount++;
    }

    Block blocks[BLOCKS];
    size_t unitSize;
    uint32_t lastBlock;
    uint32_t blockCount;
    size_t lastBlockBitsetCount;
};
