// vim: set ts=2 sw=2 et:
#include "prevalidator.h"
#include "validator.h"
#include "profiling.h"
#include <memory.h>
#include <type_traits>

using namespace std;
using namespace folly;

static const size_t VALIDATION_UNIT_COUNT = 64;

namespace Prevalidator
{
ThreadContext::ThreadContext(int _threadIndex)
: threadIndex(_threadIndex), validationQueryThreadIndex(0)
{
  validationQueueItems = new remove_reference<decltype(validationQueueItems[0])>::type[relationsEnd - relations];
#ifdef _PROFILE
  timeRelation = new ProfilingStats[relationsEnd - relations];
#endif
}

void ThreadContext::mainLoop()
{
  Parser<Prevalidator::ThreadContext> parser(*this);
  parser.parseLoop();
}

void ThreadContext::processValidationQueries(ValidationQueries *v)
{
  validationQueryThreadIndex++;
  if (validationQueryThreadIndex == THREAD_COUNT)
    validationQueryThreadIndex = 0;
  if (validationQueryThreadIndex != threadIndex)
    return;

  PROFILE_START(timeStack, timePreprocess);

  char *reader = v->queries;
  for (uint64_t i = 0;i != v->queryCount;i++)
  {
    auto q = reinterpret_cast<Query*>(reader);
    if (q->columnCount > 0)
    {
      auto &rel = relations[q->relationId];
      PROFILE_START(timeStack, timeRelation[q->relationId]);
      uint32_t rear = q->columnCount - 1;
      uint32_t maxIdx = 0;
      uint64_t maxDiff = 0;
      for (uint32_t k = 0;k <= rear; k++) {
        auto& c = q->columns[k];
        if (c.op != Query::Column::Equal)
        {
          if (rear > k) 
            swap(q->columns[k--], q->columns[rear--]);
        }
        else
        {
          uint64_t minVal = rel.historicMinValue[c.column];
          uint64_t maxVal = rel.historicMaxValue[c.column];
          uint64_t diff = maxVal - minVal;
          if (maxDiff < diff)
          {
            maxDiff = diff;
            maxIdx = k;
          } 
        }
      }
      if (maxIdx != 0)
        swap(q->columns[maxIdx], q->columns[0]);
    }

    auto &vq = validationQueueItems[q->relationId];
    if (__builtin_expect(vq.empty(), 0) || __builtin_expect(vq.back().size() >= VALIDATION_UNIT_COUNT, 0))
    {
      vq.emplace_back();
      vq.back().reserve(VALIDATION_UNIT_COUNT);
    }
    vq.back().emplace_back(v, q);
    reader += sizeof(Query) + (sizeof(Query::Column) * q->columnCount);
  }
}

void ThreadContext::processFlush(Flush *f)
{
  for (;;)
  {
    auto key = relationReadyEventCount.prepareWait();
    bool anyRelationsLeft = false;
    auto rPtr = relations;
    auto qPtr = validationQueueItems;
    for (;rPtr != relationsEnd;rPtr++, qPtr++)
    {
      if (qPtr->empty())
        continue;
      if (rPtr->readyForValidation.load())
      {
        PROFILE_START(timeStack, timeProduceTasks);
        auto &q = *qPtr;
        auto size = q.size();
        flushCounter += size;
        for (size_t i = 0;i < size;i++)
          validationQueue.write(move(q[i]));
        q.clear();
        PROFILE_END(timeStack);
      }
      else
        anyRelationsLeft = true;
    }
    if (anyRelationsLeft)
      relationReadyEventCount.wait(key);
    else
    {
      relationReadyEventCount.cancelWait();
      break;
    }
  }

  if ((flushCounter -= 1) == 0)
    flushReadySemaphore.post();
  flushCompleteSemaphore.wait();
}
} // namespace Prevalidator
