// vim: set ts=2 sw=2 et:

#include "reader.h"
#include "protocol.h"
#include <thread>
#include <memory.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <iostream>

using namespace std;

const auto readBufferSize = 4ULL * 1024 * 1024 * 1024 + 8; // 4GiB + sizeof(MessageHead)

char *readBuffer;
char *readBufferLimit;
std::atomic<char *> readBufferAvailable;
std::atomic<int> readBufferFlushEvent;
std::atomic<char *> readBufferFlushBase;
static int pageSize;
static char *readBufferMessageEnd;

folly::EventCount waitForBytesEventCount;
folly::EventCount flushConsumerCompleteEventCount;

const int READ_BUFFER_FLUSH_EVENT_OK = 0;
const int READ_BUFFER_FLUSH_EVENT_WAIT_FOR_CONSUMER = 1;
const int READ_BUFFER_FLUSH_EVENT_WAIT_FOR_PRODUCER = 2;

static void readThreadMain()
{
  char *_readBufferAvailable = readBuffer;

  pollfd pollDesc;
  pollDesc.fd = STDIN_FILENO;
  pollDesc.events = POLLIN;
  for (;;)
  {
    int pollRet = poll(&pollDesc, 1, -1);
    if (pollRet < 0)
    {
      perror("readThreadMain poll()");
      exit(1);
    }

    if (!(pollDesc.revents & POLLIN))
      return;

    int bytesAvailable;
    if (ioctl(STDIN_FILENO, FIONREAD, &bytesAvailable) < 0)
    {
      perror("readThreadMain ioctl()");
      exit(1);
    }

    if (bytesAvailable == 0)
      return;

    off_t readBufferEmptyBytes = readBufferLimit - _readBufferAvailable;
    bytesAvailable = min(min(readBufferEmptyBytes, static_cast<off_t>(bytesAvailable)), 16L * 1024 * 1024);

    char *nextReadBufferAvailable = _readBufferAvailable + bytesAvailable;

    if (read(STDIN_FILENO, _readBufferAvailable, bytesAvailable) != bytesAvailable)
    {
      perror("readThreadMain read()");
      exit(1);
    }

    _readBufferAvailable = nextReadBufferAvailable;
    for (;;)
    {
      bool shouldFlush = (_readBufferAvailable == readBufferLimit);
      char *_nextReadBufferMessageEnd = readBufferMessageEnd;
      for (;;)
      {
        if (_nextReadBufferMessageEnd + sizeof(MessageHead) > _readBufferAvailable)
          break;
        MessageHead *head = reinterpret_cast<MessageHead *>(_nextReadBufferMessageEnd);
        if (_nextReadBufferMessageEnd + sizeof(MessageHead) + head->messageLen > _readBufferAvailable)
          break;
        _nextReadBufferMessageEnd += sizeof(MessageHead) + head->messageLen;
        if (head->type == MessageHead::Flush)
        {
          shouldFlush = true;
          break;
        }
      }

      int nextFlushEvent = shouldFlush ? READ_BUFFER_FLUSH_EVENT_WAIT_FOR_CONSUMER : READ_BUFFER_FLUSH_EVENT_OK;
      int prevFlushEvent = readBufferFlushEvent.load();
      if (__builtin_expect(prevFlushEvent != nextFlushEvent, 0))
        readBufferFlushEvent.store(nextFlushEvent);

      if (readBufferMessageEnd != _nextReadBufferMessageEnd || shouldFlush)
      {
        // we notify consumers only if we have read at least one message
        readBufferMessageEnd = _nextReadBufferMessageEnd;
        readBufferAvailable.store(_nextReadBufferMessageEnd);
        waitForBytesEventCount.notifyAll();
      }

      if (prevFlushEvent == READ_BUFFER_FLUSH_EVENT_WAIT_FOR_PRODUCER)
        flushConsumerCompleteEventCount.notifyAll();

      if (!shouldFlush)
      {
        std::this_thread::yield();
        // make some of the next pages hot
        size_t currentLen = _readBufferAvailable - readBuffer;
        size_t pagesLeft = 32;
        for (currentLen += pageSize;pagesLeft > 0 && currentLen < readBufferSize;currentLen += pageSize, pagesLeft--)
        {
          char *p = readBuffer + currentLen;
          *p = 0;
        }
        break;
      }

      flushConsumerCompleteEventCount.await([](){ return readBufferFlushEvent.load() != READ_BUFFER_FLUSH_EVENT_WAIT_FOR_CONSUMER; });

      char *_readBufferFlushBase = readBufferFlushBase.load();
      auto bytesToMove = _readBufferAvailable - _readBufferFlushBase;
      memmove(readBuffer, _readBufferFlushBase, bytesToMove);
      _readBufferAvailable = readBuffer + bytesToMove;
      readBufferAvailable.store(readBuffer);
      readBufferMessageEnd = readBuffer;


      readBufferFlushEvent.store(READ_BUFFER_FLUSH_EVENT_OK);
      flushConsumerCompleteEventCount.notifyAll();
    }
  }
}

void startReaderThread()
{
  pageSize = sysconf(_SC_PAGE_SIZE);

  readBuffer = (char *)mmap(NULL, readBufferSize, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
  readBufferMessageEnd = readBuffer;
  readBufferLimit = readBuffer + readBufferSize;
  readBufferAvailable.store(readBuffer);
  readBufferFlushEvent.store(READ_BUFFER_FLUSH_EVENT_OK);
  readBufferFlushBase.store(nullptr);

  thread reader(readThreadMain);
  reader.detach();
}

