// vim: set ts=2 sw=2 et:
#include "transactor.h"
#include "validator.h"
#include "prevalidator.h"
#include <memory.h>

using namespace std;
using namespace folly;

static const size_t VALIDATION_UNIT_COUNT = 256;

namespace Transactor
{
ThreadContext::ThreadContext(int _threadIndex)
: threadIndex(_threadIndex), shouldForget(false), forgetTransactionId(0)
{
#ifdef _PROFILE
  timeRelation = new ProfilingStats[relationsEnd - relations];
#endif
}

void ThreadContext::mainLoop()
{
  Parser<Transactor::ThreadContext> parser(*this);
  parser.parseLoop();
}

inline static void transactionDryDelete(ThreadContext *self, const Transaction *t, const char *(&reader))
{
  for (uint32_t index = 0;index != t->deleteCount;++index)
  {
    auto &o = *reinterpret_cast<const TransactionOperationDelete*>(reader);

    auto &rel = relations[o.relationId];
    if (rel.ownerThreadIndex == self->threadIndex)
    {
      PROFILE_START(self->timeStack, self->timeRelation[o.relationId]);
      self->transactionDeletes.emplace_back(&o);

      rel.upcomingReservedRows += o.rowCount;
    }

    reader += sizeof(TransactionOperationDelete) + (sizeof(uint64_t) * o.rowCount);
  }
}

inline static void transactionDryInsert(ThreadContext *self, const Transaction *t, const char *(&reader))
{
  for (uint32_t index = 0;index != t->insertCount;++index)
  {
    auto &o = *reinterpret_cast<const TransactionOperationInsert*>(reader);

    auto &rel = relations[o.relationId];
    if (rel.ownerThreadIndex == self->threadIndex)
    {
      PROFILE_START(self->timeStack, self->timeRelation[o.relationId]);
      self->transactionInserts.emplace_back(&o);

      rel.upcomingReservedRows += o.rowCount;
    }

    reader += sizeof(TransactionOperationInsert) + (sizeof(uint64_t) * o.rowCount * rel.columnCount);
  }
}

inline static void buildNewHistoryRecord(Relation &rel, uint64_t xactId)
{
  if (rel.xactHistory.empty())
    rel.xactHistory.emplace_back(xactId, 0);
  else
  {
    auto &back = rel.xactHistory.back();
    rel.xactHistory.emplace_back(xactId, back.accumulatedRows + back.rows);
  }
  auto &back = rel.xactHistory.back();
  back.reservedRows = rel.upcomingReservedRows;
  back.activate(rel.columnCount);
  rel.upcomingReservedRows = 0;
}

inline static void transactionDelete(ThreadContext *self, uint64_t xactId)
{
  for (auto _o : self->transactionDeletes)
  {
    auto &o = *_o;
    PROFILE_START(self->timeStack, self->timeRelation[o.relationId]);

    auto &rel = relations[o.relationId];
    if (rel.upcomingReservedRows == 0)
    {
      // do nothing
    }
    else if (rel.upcomingReservedRows <= Relation::SMALL_ROWS)
    {
      const uint64_t *key = o.keys;
      const uint64_t *keyLimit = key + o.rowCount;
      for (;key != keyLimit;++key)
      {
        auto rowPtr = rel.activeRows.getAndRemove(*key);
        if (rowPtr == nullptr)
          continue;
        // for deleted rows, we can claim ownership here
        rel.smallPendingRows.emplace_back(xactId, rowPtr);
      }
      rel.upcomingReservedRows -= o.rowCount;
      continue;
    }
    else
    {
      buildNewHistoryRecord(rel, xactId);
    }

    auto &xact = rel.xactHistory.back();
    auto &columnarValues = xact.columnarValues;
    auto &bloomFilters = xact.bloomFilters;

    const uint64_t *key = o.keys;
    const uint64_t *keyLimit = key + o.rowCount;
    for (;key != keyLimit;++key)
    {
      auto rowPtr = rel.activeRows.getAndRemove(*key);
      if (rowPtr == nullptr)
        continue;

      auto readPos = Uint64MemoryPool::align(rowPtr);
      auto readPosEnd = readPos + rel.columnCount;
      auto writePos = columnarValues.insertPtr++;
      auto rowIndex = writePos - columnarValues.begin;
      uint32_t column = 0;
      for (;readPos != readPosEnd;)
      {
        *writePos = *readPos;
        if (bloomFilters.isActive())
        {
          bloomFilters.addElement(xact.reservedRows, rel.columnCount, rowIndex, column, *readPos);
        }
        column++;
        writePos += xact.reservedRows;
        readPos++;
      }

      rel.memoryPool.deallocate(rowPtr);
    }

    xact.rows = columnarValues.insertPtr - columnarValues.begin;
  }
  self->transactionDeletes.clear();
}

inline static void transactionInsert(ThreadContext *self, uint64_t xactId)
{
  for (auto _o : self->transactionInserts)
  {
    auto &o = *_o;
    PROFILE_START(self->timeStack, self->timeRelation[o.relationId]);

    auto &rel = relations[o.relationId];
    if (rel.upcomingReservedRows == 0)
    {
      // do nothing
    }
    else if (rel.upcomingReservedRows <= Relation::SMALL_ROWS)
    {
      auto values = o.values;
      auto valuesLimit = values + (o.rowCount * rel.columnCount);
      for (;values != valuesLimit;values += rel.columnCount)
      {
        auto readPos = values;
        auto readPosEnd = values + rel.columnCount;
        uint32_t column = 0;
        for (;readPos != readPosEnd;)
        {
          if (rel.minValue[column] > *readPos) {
            rel.minValue[column] = *readPos;
          }
          if (rel.maxValue[column] < *readPos) {
            rel.maxValue[column] = *readPos;
          }
          column++;
          readPos++;
        }
        uint64_t *rowForActiveRows = rel.memoryPool.allocate();
        uint64_t *rowForPendingRows = rel.memoryPool.allocate();
        memcpy(Uint64MemoryPool::align(rowForActiveRows), values, sizeof(uint64_t) * rel.columnCount);
        memcpy(Uint64MemoryPool::align(rowForPendingRows), values, sizeof(uint64_t) * rel.columnCount);
        rel.activeRows.set(*values, rowForActiveRows);
        rel.smallPendingRows.emplace_back(xactId, rowForPendingRows);
      }
      rel.upcomingReservedRows -= o.rowCount;
      continue;
    }
    else
    {
      buildNewHistoryRecord(rel, xactId);
    }

    auto &xact = rel.xactHistory.back();
    auto &columnarValues = xact.columnarValues;
    auto &bloomFilters = xact.bloomFilters;

    auto values = o.values;
    auto valuesLimit = values + (o.rowCount * rel.columnCount);
    for (;values != valuesLimit;values += rel.columnCount)
    {
      auto readPos = values;
      auto readPosEnd = values + rel.columnCount;
      auto writePos = columnarValues.insertPtr++;
      auto rowIndex = writePos - columnarValues.begin;
      uint32_t column = 0;
      for (;readPos != readPosEnd;)
      {
        *writePos = *readPos;
        if (rel.minValue[column] > *readPos) {
          rel.minValue[column] = *readPos;
        }
        if (rel.maxValue[column] < *readPos) {
          rel.maxValue[column] = *readPos;
        }
        if (bloomFilters.isActive())
        {
          bloomFilters.addElement(xact.reservedRows, rel.columnCount, rowIndex, column, *readPos);
        }
        column++;
        writePos += xact.reservedRows;
        readPos++;
      }
      uint64_t *row = rel.memoryPool.allocate();
      memcpy(Uint64MemoryPool::align(row), values, sizeof(uint64_t) * rel.columnCount);
      rel.activeRows.set(*values, row);
    }

    xact.rows = columnarValues.insertPtr - columnarValues.begin;
  }
  self->transactionInserts.clear();
}

void ThreadContext::processTransaction(Transaction *t)
{
  PROFILE_START(timeStack, timeTransaction);

  const char* reader = t->operations;

  PROFILE_START(timeStack, timeTransactionDry);
  transactionDryDelete(this, t, reader);
  transactionDryInsert(this, t, reader);
  PROFILE_END(timeStack);

  PROFILE_START(timeStack, timeTransactionDelete);
  transactionDelete(this, t->transactionId);
  PROFILE_END(timeStack);

  PROFILE_START(timeStack, timeTransactionInsert);
  transactionInsert(this, t->transactionId);
  PROFILE_END(timeStack);
}

static void buildSmallXactHistory(Transactor::ThreadContext *self)
{
  PROFILE_START(self->timeStack, self->timeBuildSmall);
  for (Relation *ptr = relations;ptr != relationsEnd;ptr++)
  {
    Relation &rel = *ptr;
    if (rel.ownerThreadIndex != self->threadIndex)
      continue;
    PROFILE_START(self->timeStack, self->timeRelation[ptr - relations]);

    if (rel.smallPendingRows.empty())
      continue;

    uint64_t accumulatedRows = 0;
    if (!rel.smallXactHistory.empty())
      accumulatedRows = rel.smallXactHistory.back().accumulatedRows + rel.smallXactHistory.back().rows;

    rel.smallXactHistory.emplace_back(
      rel.smallPendingRows.front().first,
      rel.smallPendingRows.back().first,
      accumulatedRows
    );
    auto &xact = rel.smallXactHistory.back();
    xact.reservedRows = rel.smallPendingRows.size();
    xact.rows = rel.smallPendingRows.size();
    xact.activate(rel.columnCount);

    auto &columnarValues = xact.columnarValues;
    auto &bloomFilters = xact.bloomFilters;
    auto &xactIds = xact.xactIds;
    auto &xactIdToRowIndex = xact.xactIdToRowIndex;

    auto begin = rel.smallPendingRows.data();
    auto end = begin + rel.smallPendingRows.size();
    uint64_t rowIndex = 0;
    uint64_t lastXactId = -1;
    for (auto it = begin;it != end;it++, rowIndex++)
    {
      uint64_t *unalignedRow = it->second;
      uint64_t *alignedRow = Uint64MemoryPool::align(unalignedRow);
      auto readPos = alignedRow;
      auto readPosEnd = alignedRow + rel.columnCount;
      auto writePos = columnarValues.insertPtr++;
      uint32_t column = 0;
      xactIds[rowIndex] = it->first;
      if (rowIndex == 0 || lastXactId != it->first)
      {
        xactIdToRowIndex.emplace_back(it->first, rowIndex);
        lastXactId = it->first;
      }
      for (;readPos != readPosEnd;)
      {
        *writePos = *readPos;
        if (bloomFilters.isActive())
        {
          bloomFilters.addElement(xact.reservedRows, rel.columnCount, rowIndex, column++, *readPos);
        }
        writePos += xact.reservedRows;
        readPos++;
      }
      rel.memoryPool.deallocate(unalignedRow);
    }
    rel.smallPendingRows.clear();
  }
}

void ThreadContext::processFlush(Flush *f)
{
  buildSmallXactHistory(this);

  for (Relation *ptr = relations;ptr != relationsEnd;ptr++)
  {
    Relation &rel = *ptr;
    if (rel.ownerThreadIndex != threadIndex)
      continue;
    rel.readyForValidation = true;
  }
  relationReadyEventCount.notifyAll();

  if ((flushCounter -= 1) == 0)
    flushReadySemaphore.post();
  flushCompleteSemaphore.wait();

  if (shouldForget)
  {
    forget(forgetTransactionId);
    shouldForget = false;
  }
}

void ThreadContext::processForget(Forget *f)
{
  shouldForget = true;
  forgetTransactionId = f->transactionId;
}

void ThreadContext::forget(uint64_t forgetTransactionId)
{
  PROFILE_START(timeStack, timeForget);

  for (Relation *ptr = relations;ptr != relationsEnd;ptr++)
  {
    Relation &rel = *ptr;
    if (rel.ownerThreadIndex != threadIndex)
      continue;
    PROFILE_START(timeStack, timeRelation[ptr - relations]);
    auto it = upper_bound(rel.xactHistory.begin(), rel.xactHistory.end(), forgetTransactionId,
      [](uint64_t transactionId, decltype(*rel.xactHistory.begin()) rec) {
        return transactionId < rec.xactId;
      });
    decltype(rel.xactHistory.rend()) rit(it);
    for (;rit != rel.xactHistory.rend() && !rit->isInactive();rit++)
      rit->deactivate();

    auto sit = upper_bound(rel.smallXactHistory.begin(), rel.smallXactHistory.end(), forgetTransactionId,
      [](uint64_t transactionId, decltype(*rel.smallXactHistory.begin()) rec) {
        return transactionId < rec.endXactId;
      });
    decltype(rel.smallXactHistory.rend()) rsit(sit);
    for (;rsit != rel.smallXactHistory.rend() && !rsit->isInactive();rsit++)
      rsit->deactivate();
  }
}
} // namespace Transactor
