// vim: set ts=2 sw=2 et:
#include "validator.h"

#include <memory.h>
#include <immintrin.h>

using namespace std;
using BloomFilters = XactHistoryRecord::BloomFilters;

folly::MPMCQueue<folly::fbvector<Validator::ValidationSubtask>> validationQueue(131072);

namespace Validator
{
ThreadContext::ThreadContext()
{
  bitColumnIntersectionReservedRows = 16384;
  BloomFilters::allocateMemory(&bitColumnIntersection, bitColumnIntersectionReservedRows, 1);
#ifdef _PROFILE
  timeRelation = new ProfilingStats[relationsEnd - relations];
#endif
}

void ThreadContext::mainLoop()
{
  folly::fbvector<ValidationSubtask> queries;
  while (true)
  {
    validationQueue.blockingRead(queries);

    PROFILE_START(this->timeStack, this->timeValidation);
    for (auto &query : queries)
      validation(query);
    PROFILE_END(this->timeStack);

    if ((flushCounter -= 1) == 0)
      flushReadySemaphore.post();
  }
}

inline static bool validationQueryColumnPreprocess(ThreadContext *self, Query *q)
{
  self->columnHashedValue.resize(q->columnCount);
  self->columnHashedValueCompute = 0;

  auto &rel = relations[q->relationId];
  uint32_t k = 0;
  for (k = 0;k < q->columnCount && q->columns[k].op == Query::Column::Equal;k++)
  {
    auto& c = q->columns[k];
    uint64_t minVal = rel.minValue[c.column];
    uint64_t maxVal = rel.maxValue[c.column];
    if (minVal > c.value || maxVal < c.value)
      return false;
  }
  auto equalCount = k;
  for (;k < q->columnCount;k++)
  {
    auto& c = q->columns[k];
    uint64_t minVal = rel.minValue[c.column];
    uint64_t maxVal = rel.maxValue[c.column];
    switch (c.op)
    {
      case Query::Column::Less:
        if (c.value <= minVal) return false; break;
      case Query::Column::LessOrEqual:
        if (c.value < minVal) return false; break;
      case Query::Column::Greater:
        if (c.value >= maxVal) return false; break;
      case Query::Column::GreaterOrEqual:
        if (c.value > maxVal) return false; break;
      default:
        break;
    }
  }

  self->columnsBegin = q->columns;
  self->columnsEqualEnd = q->columns + equalCount;
  self->columnsEnd = q->columns + q->columnCount;

  return true;
}

template<typename T, bool IS_SMALL>
struct ValidateHelper
{
  static void buildResultSet(T hit, uint64_t from, uint64_t to, folly::fbvector<uint64_t> &resultSet);
  static void getXactIdRange(T hit, uint64_t from, uint64_t to, uint64_t& begin, uint64_t& end);
};

template<typename T>
struct ValidateHelper<T, false>
{
  static void buildResultSet(T hit, uint64_t from, uint64_t to, folly::fbvector<uint64_t> &resultSet)
  {
    for (uint64_t i = 0;i < hit->rows;i++)
      resultSet.emplace_back(i);
  }
  static void getXactIdRange(T hit, uint64_t from, uint64_t to, uint64_t& begin, uint64_t& end) {
    begin = 0;
    end = hit->rows;
  }
};

template<typename T>
struct ValidateHelper<T, true>
{
  static void getXactIdRange(T hit, uint64_t from, uint64_t to, uint64_t& begin, uint64_t& end) {
    auto pBegin = lower_bound(hit->xactIdToRowIndex.begin(), hit->xactIdToRowIndex.end(), from,
        [](decltype(*hit->xactIdToRowIndex.begin()) rec, uint64_t from) {
        return rec.first < from;
        });
    auto pEnd = upper_bound(pBegin, hit->xactIdToRowIndex.end(), to,
        [](uint64_t to, decltype(*hit->xactIdToRowIndex.begin()) rec) {
        return to < rec.first;
        });
    begin = pBegin->second;
    end = (pEnd == hit->xactIdToRowIndex.end() ? hit->rows : pEnd->second);
  }

  static void buildResultSet(T hit, uint64_t from, uint64_t to, folly::fbvector<uint64_t> &resultSet)
  {
    auto pBegin = lower_bound(hit->xactIdToRowIndex.begin(), hit->xactIdToRowIndex.end(), from,
      [](decltype(*hit->xactIdToRowIndex.begin()) rec, uint64_t from) {
        return rec.first < from;
      });
    auto pEnd = upper_bound(pBegin, hit->xactIdToRowIndex.end(), to,
      [](uint64_t to, decltype(*hit->xactIdToRowIndex.begin()) rec) {
        return to < rec.first;
      });
    auto vBegin = pBegin->second;
    auto vEnd = (pEnd == hit->xactIdToRowIndex.end() ? hit->rows : pEnd->second);
    for (auto v = vBegin;v != vEnd;v++)
      resultSet.emplace_back(v);
  }
};

template<typename T, bool IS_SMALL>
static bool validateAgainstContainer(Validator::ThreadContext *self, T begin, T end, const ValidationQueries *v, const Query *q, uint64_t from, uint64_t to)
{
  bool result = false;
  auto modulus = BloomFilters::bitsPerColumn(relations[q->relationId].columnCount) - 1;

  for (auto hit = begin;hit != end;hit++)
  {
    {
      auto &columnarValues = hit->columnarValues;
      auto &bloomFilters = hit->bloomFilters;

      if (!bloomFilters.isActive())
      {
        PROFILE_START(self->timeStack, self->timeSmallValidation);
        uint64_t vBegin;
        uint64_t vEnd;
        ValidateHelper<T, IS_SMALL>::getXactIdRange(hit, from, to, vBegin, vEnd);
        bool any = false;
        self->resultSet.clear();

        if (self->columnsEqualEnd != self->columnsBegin) {
          // equal
          auto cit = self->columnsBegin;
          uint64_t *colBegin = (columnarValues.begin + cit->column * hit->reservedRows);
          uint64_t *colEnd = colBegin + vEnd;
          for (auto *col = colBegin + vBegin;col != colEnd;col++)
          {
            if (*col == cit->value) {
              self->resultSet.emplace_back(col - colBegin);
            }
          }
          for (auto id : self->resultSet)
          {
            for (auto cit = self->columnsBegin+1;cit != self->columnsEqualEnd;cit++) {
              uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows); 
              if (col[id] != cit->value) 
                goto small_validation_failed;
            }
            for (auto cit = self->columnsEqualEnd;cit != self->columnsEnd;cit++) {
              uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows); 
              switch (cit->op) {
                case Query::Column::Less:
                  if (col[id] >= cit->value) goto small_validation_failed; break;
                case Query::Column::LessOrEqual:
                  if (col[id] > cit->value) goto small_validation_failed; break;
                case Query::Column::Greater:
                  if (col[id] <= cit->value) goto small_validation_failed; break;
                case Query::Column::GreaterOrEqual:
                  if (col[id] < cit->value) goto small_validation_failed; break;
                case Query::Column::NotEqual:
                  if (col[id] == cit->value) goto small_validation_failed; break;
                default:
                  break;
              }
            }
            any = true;
            break;

small_validation_failed:
            ; // do nothing
          }
        } else {
          for (size_t id = vBegin; id < vEnd; id++) 
          {
            for (auto cit = self->columnsBegin;cit != self->columnsEqualEnd;cit++) {
              uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows); 
              if (col[id] != cit->value) 
                goto small_validation_failed2;
            }
            for (auto cit = self->columnsEqualEnd;cit != self->columnsEnd;cit++) {
              uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows); 
              switch (cit->op) {
                case Query::Column::Less:
                  if (col[id] >= cit->value) goto small_validation_failed2; break;
                case Query::Column::LessOrEqual:
                  if (col[id] > cit->value) goto small_validation_failed2; break;
                case Query::Column::Greater:
                  if (col[id] <= cit->value) goto small_validation_failed2; break;
                case Query::Column::GreaterOrEqual:
                  if (col[id] < cit->value) goto small_validation_failed2; break;
                case Query::Column::NotEqual:
                  if (col[id] == cit->value) goto small_validation_failed2; break;
                default:
                  break;
              }
            }
            any = true;
            break;

small_validation_failed2:
            ; // do nothing
          }
        }

        if (any)
        {
          result = true;
          break;
        }
        else
          goto no_record_in_xactHistory;
      }

      self->resultSet.clear();

      if (self->columnsBegin != self->columnsEqualEnd)
      {
        if (hit->reservedRows > self->bitColumnIntersectionReservedRows)
        {
          self->bitColumnIntersectionReservedRows = hit->reservedRows * 3 / 2;
          free(self->bitColumnIntersection);
          BloomFilters::allocateMemory(&self->bitColumnIntersection, self->bitColumnIntersectionReservedRows, 1);
        }
        auto elementsPerBitColumn = BloomFilters::elementsPerBitColumn(hit->reservedRows);
        auto bytesPerBitColumn = BloomFilters::bytesPerBitColumn(hit->reservedRows);

        PROFILE_START(self->timeStack, self->timeIntersection);
        uint32_t k = 0;
        for (auto cit = self->columnsBegin;cit != self->columnsEqualEnd;cit++, k++)
        {
          if (k >= self->columnHashedValueCompute)
          {
            self->columnHashedValueCompute = k + 1;
            self->columnHashedValue[k] = (BloomFilters::hash(cit->column, cit->value) & modulus);
          }
          auto hashed = self->columnHashedValue[k];
          auto bloomFiltersBitColumn = bloomFilters.bloomFilters + elementsPerBitColumn * hashed;

          if (__builtin_expect(cit == self->columnsBegin, 0))
          {
            // first run
            memcpy(self->bitColumnIntersection, bloomFiltersBitColumn, bytesPerBitColumn);
          }
          else
          {
            float *pPtr = reinterpret_cast<float *>(self->bitColumnIntersection);
            float *qPtr = reinterpret_cast<float *>(bloomFiltersBitColumn);

            auto qqwordsPerBitColumn = elementsPerBitColumn >> 2;
            __m256 over = _mm256_setzero_ps();
            for (size_t j = 0;j < qqwordsPerBitColumn;j++)
            {
              __m256 v = _mm256_and_ps(_mm256_load_ps(pPtr), _mm256_load_ps(qPtr));
              over = _mm256_or_ps(over, v);
              _mm256_store_ps(pPtr, v);
              pPtr += 8;
              qPtr += 8;
            }
            __m256i overi = _mm256_castps_si256(over);
            if (_mm256_testz_si256(overi, overi))
              goto no_record_in_xactHistory;
          }
        }
        PROFILE_END(self->timeStack);

        PROFILE_START(self->timeStack, self->timeExtraction);
        uint64_t *ptr = self->bitColumnIntersection;
        uint64_t rowIndex = 0;

        bool endXactRange = false;

        uint64_t vBegin;
        uint64_t vEnd;
        ValidateHelper<T, IS_SMALL>::getXactIdRange(hit, v->from, v->to, vBegin, vEnd);

        for (size_t i = 0;i < elementsPerBitColumn;i++)
        {
          uint64_t element = *(ptr++);
          for (;element;)
          {
            int zeroes = __builtin_ctzll(element);
            if (vBegin <= rowIndex + zeroes && rowIndex + zeroes < vEnd) {
              self->resultSet.push_back(rowIndex + zeroes);
            }
            else if (rowIndex + zeroes >= vEnd) {
              endXactRange = true;
              break;
            }
            element ^= (1ULL << zeroes);
          }
          if (endXactRange)
            break;
          rowIndex += 64;
        }
        if (self->resultSet.empty())
          goto no_record_in_xactHistory;
        PROFILE_END(self->timeStack);
      }
      else
        ValidateHelper<T, IS_SMALL>::buildResultSet(hit, from, to, self->resultSet);

      PROFILE_START(self->timeStack, self->timeBloomFilterValidation);

      folly::fbvector<uint64_t> tmpFlags;
      for (auto cit = self->columnsBegin; cit != self->columnsEqualEnd; cit++) {
        uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows);
        for (auto x : self->resultSet )
          if (col[x] == cit->value)
            tmpFlags.push_back(x);
        if (tmpFlags.empty()) 
          goto no_record_in_xactHistory; 
        swap(self->resultSet, tmpFlags);
        tmpFlags.clear();
      }

      for (auto cit = self->columnsEqualEnd;cit != self->columnsEnd;cit++) {
        uint64_t *col = (columnarValues.begin + cit->column * hit->reservedRows); 
        for (auto x : self->resultSet) {
          switch (cit->op) {
            case Query::Column::Less:
              if (col[x] < cit->value)
                tmpFlags.push_back(x);
              break;
            case Query::Column::LessOrEqual:
              if (col[x] <= cit->value)
                tmpFlags.push_back(x);
              break;
            case Query::Column::Greater:
              if (col[x] > cit->value)
                tmpFlags.push_back(x);
              break;
            case Query::Column::GreaterOrEqual:
              if (col[x] >= cit->value)
                tmpFlags.push_back(x);
              break;
            case Query::Column::NotEqual:
              if (col[x] != cit->value)
                tmpFlags.push_back(x);
              break;
            default:
              break;
          }
        }
        if (tmpFlags.empty()) 
          goto no_record_in_xactHistory; 
        swap(self->resultSet, tmpFlags);
        tmpFlags.clear();
      }
      PROFILE_END(self->timeStack);

      result = true;
      break;
    }

no_record_in_xactHistory:
    ; // do nothing
  }

  return result;
}

void ThreadContext::validation(ValidationSubtask &subtask)
{
  auto v = subtask.v;
  auto q = subtask.q;
  auto from = v->from;
  auto to = v->to;
  auto validationOffset = v->validationId - firstUnknownValidationId;
  if (validationOffset >= validationResults.size())
    validationResults.resize(validationOffset + 1);
  auto &result = validationResults[validationOffset];
  if (result == VALIDATION_TRUE)
    return;
  result = VALIDATION_FALSE;

  PROFILE_START(timeStack, timeRelation[q->relationId]);

  if (!validationQueryColumnPreprocess(this, q))
    return;

  auto &rel = relations[q->relationId];
  auto xactHistoryBegin = lower_bound(rel.xactHistory.data(), rel.xactHistory.data() + rel.xactHistory.size(), from,
      [](decltype(*rel.xactHistory.begin()) rec, uint64_t from) {
      return rec.xactId < from;
      });
  auto xactHistoryEnd = upper_bound(xactHistoryBegin, rel.xactHistory.data() + rel.xactHistory.size(), v->to,
      [](uint64_t to, decltype(*rel.xactHistory.begin()) rec) {
      return to < rec.xactId;
      });
  if (validateAgainstContainer<decltype(xactHistoryBegin), false>(this, xactHistoryBegin, xactHistoryEnd, v, q, from, to))
  {
    result = VALIDATION_TRUE;
    return;
  }

  auto smallXactHistoryBegin = lower_bound(rel.smallXactHistory.data(), rel.smallXactHistory.data() + rel.smallXactHistory.size(), from,
      [](decltype(*rel.smallXactHistory.begin()) rec, uint64_t from) {
      return rec.endXactId < from;
      });
  auto smallXactHistoryEnd = upper_bound(smallXactHistoryBegin, rel.smallXactHistory.data() + rel.smallXactHistory.size(), v->to,
      [](uint64_t to, decltype(*rel.smallXactHistory.begin()) rec) {
      return to < rec.xactId;
      });
  if (validateAgainstContainer<decltype(smallXactHistoryBegin), true>(this, smallXactHistoryBegin, smallXactHistoryEnd, v, q, from, to))
  {
    result = VALIDATION_TRUE;
    return;
  }
}
} // namespace Validator
