#pragma once

#include <cstdint>
#include <cstdlib>
#include <cstring>
#include "ptr-hash-map.h"
#include "folly/FBVector.h"

struct XactHistoryRecord
{
    struct ColumnarValues
    {
        uint64_t *begin;
        uint64_t *insertPtr;

        inline ColumnarValues()
            : begin(nullptr), insertPtr(nullptr)
        { }

        inline void activate(uint32_t columnCount, XactHistoryRecord &xact)
        {
            begin = new uint64_t[columnCount * xact.reservedRows];
            insertPtr = begin;
        }

        inline void deactivate()
        {
            delete[] begin;
            begin = nullptr;
            insertPtr = nullptr;
        }
    };

    struct BloomFilters
    {
        // TODO: PLAY WITH VALUES
        static constexpr uint32_t BITS_PER_COLUMN_FACTOR = 9;
        static constexpr uint32_t MINIMAL_RESERVED_ROWS = 128;

        // Total elements: 
        //  ceil(reservedRows / 256) * 4 * columnCount * BITS_PER_COLUMN;
        // Conceptually each row has an own instance of bloom filter with size (columnCount * BITS_PER_COLUMN), but they are stored bit-column-wise.
        // These values are aligned in 256-bit boundary.
        uint64_t *bloomFilters;
        uint64_t _bitsPerColumn;
        uint64_t modulus;

        inline BloomFilters()
            : bloomFilters(nullptr), _bitsPerColumn(0), modulus(0)
        { }

        inline static size_t bitsPerColumn(uint32_t columnCount) {
            return 1ULL << (64 - __builtin_clzll(static_cast<uint64_t>(columnCount) * BITS_PER_COLUMN_FACTOR - 1));
        }

        inline static size_t elementsPerBitColumn(uint64_t reservedRows)
        {
            return (reservedRows + 255) / 256 * 256 / 8 / sizeof(uint64_t);
        }

        inline static size_t bytesPerBitColumn(uint64_t reservedRows)
        {
            return (reservedRows + 255) / 256 * 256 / 8;
        }

        inline static void allocateMemory(uint64_t **ptr, uint64_t reservedRows, size_t bitColumnCount)
        {
            size_t totalBytes = bytesPerBitColumn(reservedRows) * bitColumnCount;
            if (posix_memalign(reinterpret_cast<void **>(ptr), 256 / 8, totalBytes))
                perror("BloomFilters posix_memalign()");
        }

        inline static void allocateMemoryZerofill(uint64_t **ptr, uint64_t reservedRows, size_t bitColumnCount)
        {
            size_t totalBytes = bytesPerBitColumn(reservedRows) * bitColumnCount;
            if (posix_memalign(reinterpret_cast<void **>(ptr), 256 / 8, totalBytes))
                perror("BloomFilters posix_memalign()");
            memset(*ptr, 0, totalBytes);
        }

        inline bool isActive()
        {
            return bloomFilters != nullptr;
        }

        inline void activate(uint32_t columnCount, XactHistoryRecord &xact)
        {
            if (xact.reservedRows >= MINIMAL_RESERVED_ROWS)
            {
                _bitsPerColumn = bitsPerColumn(columnCount);
                modulus = _bitsPerColumn - 1;
                allocateMemoryZerofill(&bloomFilters, xact.reservedRows, _bitsPerColumn);
            }
        }

        inline void deactivate()
        {
            free(bloomFilters);
            bloomFilters = nullptr;
        }

        inline static uint32_t hash(uint32_t columnIndex, uint64_t value)
        {
            return PtrHashMap<void *>::hash(value ^ columnIndex);
        }

        inline void addBit(uint64_t reservedRows, uint64_t rowIndex, uint32_t bitColumnIndex)
        {
            uint64_t bitIndex = bytesPerBitColumn(reservedRows) * 8 * bitColumnIndex + rowIndex;
            uint64_t &value = bloomFilters[bitIndex >> 6];
            value = (value | (1ULL << (bitIndex & 63)));
        }

        inline void addElement(uint64_t reservedRows, uint32_t columnCount, uint64_t rowIndex, uint32_t columnIndex, uint64_t value)
        {
            addBit(reservedRows, rowIndex, hash(columnIndex, value) & modulus);
        }
    };

    uint64_t xactId;
    uint64_t rows;
    uint64_t reservedRows;
    uint64_t accumulatedRows;
    ColumnarValues columnarValues;
    BloomFilters bloomFilters;

    inline XactHistoryRecord(uint64_t _xactId, uint64_t _accumulatedRows)
        : xactId(_xactId), rows(0), reservedRows(0), accumulatedRows(_accumulatedRows)
    { }

    inline bool isAppendable(uint64_t xactId)
    {
        return this->xactId == xactId && this->columnarValues.begin == nullptr;
    }

    inline bool isInactive()
    {
        return this->columnarValues.begin == nullptr;
    }

    inline void activate(uint32_t columnCount)
    {
        columnarValues.activate(columnCount, *this);
        bloomFilters.activate(columnCount, *this);
    }

    inline void deactivate()
    {
        columnarValues.deactivate();
        bloomFilters.deactivate();
    }
};

struct SmallXactHistoryRecord : public XactHistoryRecord
{
    uint64_t endXactId;
    uint64_t *xactIds;
    folly::fbvector<std::pair<uint64_t, uint64_t>> xactIdToRowIndex;

    inline SmallXactHistoryRecord(uint64_t _xactId, uint64_t _endXactId, uint64_t _accumulatedRows)
        : XactHistoryRecord(_xactId, _accumulatedRows), endXactId(_endXactId)
    { }

    inline void activate(uint32_t columnCount)
    {
        xactIds = new uint64_t[reservedRows];
        XactHistoryRecord::activate(columnCount);
    }

    inline void deactivate()
    {
        delete[] xactIds;
        XactHistoryRecord::deactivate();
    }
};
